/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include <linux/kthread.h>
#include <linux/wait.h>
#include "replayk.h"
#include "replay_timing.h"
#include "prefetch.h"
#include "fist.h"
#include "debug.h"

extern int fist_debug_var;

static atomic_t terminate;
int init_thread_loc = 0;

struct replay_command* fetch_cmd(struct replay* r)
{
	/* play only PAGE_SIZE/sizeof(cmd) commands by now */
	return r->cur_cmd < r->cmd_num ? &r->srcmdp[r->cur_cmd++] : NULL;
}
//The below macro has been modified to also print the actual command added by tracefs that errored out.
#define REPLAY_CMD(replay_cmd) 	do { \
					err = replay_cmd(scmdp, r); \
					if (err) \
					{\
						printk("Relcmd %d Op %d returned errro %d",relcmd, scmdp->command,err);\
						goto out; \
					}\
				} while (0);

#if MULTITHREADING
int process_replay_command(struct replay *r,struct replay_command *scmdp)
{
  int err = 0;

  print_entry_location();

  /*
   * If these asserts fail,that means the user-level compiler either has a bug
   * or the replay binary needs to be recompiled.
   */
#if REPLAY_READS
  ASSERT(scmdp->command != OP_READPAGE && scmdp->command != OP_READPAGE);
#else
  ASSERT(scmdp->command != OP_READ);
#endif

  switch (scmdp->command) {
#if PREFETCH
	case OP_LOAD:		REPLAY_CMD(replay_prefetch_load);
				goto out;
	case OP_UNLOAD:		REPLAY_CMD(replay_prefetch_unload);
				goto out;
#else
        case OP_LOAD:		REPLAY_CMD(replay_load);
				goto out;
	case OP_UNLOAD:		REPLAY_CMD(replay_unload);
	                        goto out;
#endif
#if MULTITREADHING
	case OP_CREATE_THREAD:	REPLAY_CMD(replay_create_thread);
				goto out;
#endif
	case OP_CREATE:		REPLAY_CMD(replay_create);
				break;
	case OP_LINK:		REPLAY_CMD(replay_link);
				break;
	case OP_SYMLINK:	REPLAY_CMD(replay_symlink);
				break;
	case OP_RENAME:	        REPLAY_CMD(replay_rename);
				break;
	case OP_OPEN:		REPLAY_CMD(replay_open);
				break;
	case OP_LOOKUP:		REPLAY_CMD(replay_lookup);
				break;
        case OP_FLUSH:		REPLAY_CMD(replay_flush);
				break;
	case OP_RELEASE:	REPLAY_CMD(replay_release);
				break;
	case OP_WRITE:		REPLAY_CMD(replay_write);
				break;
	case OP_READ:		REPLAY_CMD(replay_read);
				break;
	case OP_MKDIR:		REPLAY_CMD(replay_mkdir);
				break;
	case OP_READPAGE:	REPLAY_CMD(replay_readpage);
				break;
	case OP_READPAGES:	REPLAY_CMD(replay_readpages);
                    		break;
	case OP_RMDIR:          REPLAY_CMD(replay_rmdir);
				break;
	case OP_COMMIT_WRITE:	REPLAY_CMD(replay_commit_write);
				break;
	/*The sync page operation tries to call find_get_page to fetch a
      page from the mapping at a particular offset. During replay it may
      so happen that the pdflush kernel thread has already woken up and
      flushes the corresponding dirty pages of this mapping and this page is
      no longer dirty. As a result the function find_get_page returns a NULL
      page. This can happen due to the difference in system load at the time
      of tracing and replaying. Hence the replay_sync_page will try to find the
	  corresponding page and only if found flushes it to disk, else quitely
	  returns without any error.*/
	case OP_SYNC_PAGE:	REPLAY_CMD(replay_sync_page);
				break;
	case OP_UNLINK:		REPLAY_CMD(replay_unlink);
				break;
	case OP_D_RELEASE:      REPLAY_CMD(replay_d_release);
				break;
	case OP_READDIR:	REPLAY_CMD(replay_readdir);
				break;
 	case OP_WRITEPAGE:      REPLAY_CMD(replay_writepage);
                                break;
        case OP_DGET:           REPLAY_CMD(replay_dget);
	                        break;
        case OP_DPUT:           REPLAY_CMD(replay_dput);
                                break;
      	case OP_MKNOD:          REPLAY_CMD(replay_mknod);
                                break;
      	case OP_SETATTR:        REPLAY_CMD(replay_setattr);
                                break;
      	case OP_FSYNC:          REPLAY_CMD(replay_fsync);
                                break;
      	case OP_LLSEEK:         REPLAY_CMD(replay_llseek);
                                break;
      	case OP_PERMISSION:     REPLAY_CMD(replay_permission);
                                break;
	default:
		printk(KERN_WARNING"Unknown replay command encoutered, command = %d - aborting\n",scmdp->command);
		err = -EINVAL;
		goto out;
  }

 out:
  print_exit_location();
  return err;
}

void wakeup_next_thread(struct replay *r,int thread_loc,kthread_data_t *cur_kthread_data)
{
  kthread_data_t* new_kthread_data;
  print_entry_location();

  new_kthread_data = r->ratp[thread_loc].u.kthread_data;

  //Wakeup the next thread when schedule() is called
  down(new_kthread_data->sem);
  atomic_set(new_kthread_data->suspend,0);
  {
    wait_queue_t wait_queue;
    wait_queue.task = new_kthread_data->task;
    default_wake_function(&wait_queue,TASK_INTERRUPTIBLE,1,NULL);
  }
  //Suspend this thread the next time schedule() is called
  atomic_set(cur_kthread_data->suspend,1);

  print_exit_location();
  return;
}

int thread_manager(kthread_data_t *kthread_data)
{
  int err = 0;
  int thread_loc = 0;
  int op;
  struct replay *r;
  struct res_alloc_table* ratp;
  print_entry_location();

#if MULTITHREADING
  if(atomic_read(&terminate)) {
    //Previously executed command failed
    err = atomic_read(&terminate);
    goto out;
  }
#endif

  ASSERT(kthread_data);
  r = kthread_data->r;
  ratp = r->ratp;
  kthread_data->srcmdp = fetch_cmd(r);

#if !MULTITHREADING
  //Skip all create thread ops
  while(kthread_data->srcmdp != NULL) {
    if(kthread_data->srcmdp->command == OP_CREATE_THREAD)
      kthread_data->srcmdp = fetch_cmd(r);
    else
      break;
  }
  if(kthread_data->srcmdp == NULL)
    goto out;
#else
  if(kthread_data->srcmdp == NULL) {
    //finished replaying, put this thread to sleep and wakeup the initial thread
    if(init_thread_loc != kthread_data->thread_loc) {
      kthread_data_t* init_kthread_data;
      wait_queue_t wait_queue;

      init_kthread_data = r->ratp[init_thread_loc].u.kthread_data;
      wait_queue.task = init_kthread_data->task;
      init_kthread_data->terminate = 1;
      init_kthread_data->task = TASK_RUNNING;

      default_wake_function(&wait_queue,TASK_INTERRUPTIBLE,1,NULL);
      current->state = TASK_STOPPED; //current thread = STOP!
      schedule();
    }
    goto out;
  }
  thread_loc = kthread_data->srcmdp->thread_loc;
  op = kthread_data->srcmdp->command;
#endif

#if PREFETCH
  prefetch_readahead(r,r->data);
#endif

  //Output '.' every 2500 commands
   if(!(r->cur_cmd % 2500))
     printk(".");

#if MULTITHREADING
  //Decide whether to wakeup another thread
  if(r->cur_cmd+1 < r->cmd_num) {
    struct replay_command *next_cmd = &r->srcmdp[r->cur_cmd];
    int next_thread_loc = next_cmd->thread_loc;

    if(next_thread_loc != 0 && thread_loc != 0 &&
       next_thread_loc != thread_loc) {
      while(!r->ratp[next_cmd->thread_loc].u.kthread_data->initialized)
	schedule();
      wakeup_next_thread(kthread_data->r,next_thread_loc,kthread_data);
    }
  }
#endif

 out:
  print_exit_location();
  return err;
}

int thread_execute(void* arg)
{
  int err = 0;
  int thread_loc = 0;
  int first_thread_initialized = 0;
  struct replay *r;
  struct replay_command *scmdp;
  struct res_alloc_table *ratp;
  kthread_data_t *kthread_data;

  print_entry_location();

  if(init_thread_loc)
    first_thread_initialized = 1;

  ASSERT(arg);
  kthread_data = (kthread_data_t*)arg;

#if MULTITHREADING
  //Get Thread location
  thread_loc = kthread_data->thread_loc;

  //Initial Setup
  if(thread_loc != 0) {
    kthread_data->initialized = 1;
    current->state = TASK_INTERRUPTIBLE;
    schedule();
  }
#endif

  r = kthread_data->r;
  ratp = r->ratp;

  while(!err) {
#if MULTITHREADING
    if(thread_loc == 0) {
      //This is the initial thread, no create_thread op has been called yet
      if(init_thread_loc) {
	//Initial thread has a HOME!
	thread_loc = init_thread_loc;

	//Set the new kthread_data object - created by replay_create_thread()
	kthread_data = ratp[thread_loc].u.kthread_data;
      }
    }
    else {
      up(kthread_data->sem);
      if(atomic_read(kthread_data->suspend)) {
	current->state = TASK_INTERRUPTIBLE;
	schedule(); //Put the thread to sleep
      }
    }

    if(kthread_data->terminate)
      goto out;
#endif

    err = thread_manager(kthread_data);
    scmdp = kthread_data->srcmdp;
    if(scmdp == NULL)
      goto out; //Done replaying
    if(err)
      goto out; //Another thread signaled failure

#if MULTITHREADING
    //Make sure the correct thread is running!
    if(scmdp->thread_loc != 0 && thread_loc != 0)
      ASSERT(scmdp->thread_loc == thread_loc);
#endif

#if REPLAY_TIMING
    if(scmdp->command != OP_CREATE_THREAD)
      replay_timing_pre_op(scmdp,REPLAY_TIMING_ELAPSED,REPLAY_TIMING_SPEED);
#endif
    err = process_replay_command(r,scmdp);
/* #if REPLAY_TIMING
  if(scmdp->command != OP_CREATE_THREAD)
    replay_timing_post_op();
#endif */

#if MULTITHREADING
    if(scmdp->command == OP_CREATE_THREAD && first_thread_initialized) {
      //Wake up the next thread
      while(!r->ratp[scmdp->ret].u.kthread_data->initialized)
	schedule();
      wakeup_next_thread(r,scmdp->ret,kthread_data);
    }
    else
      first_thread_initialized = 1;

    if(err)
      atomic_set(&terminate,err); //Flag failure
#endif
  }

 out:
  print_exit_location();
  return err;
}

/* replay captured items in a loop */
int do_replay(struct replay* r)
{
	int err = 0;
	int cmd = -1;
	struct semaphore *sem;
	atomic_t *suspend;
	kthread_data_t *kthread_data;

	print_entry_location();

	atomic_set(&terminate,0);
	r->cur_cmd = 0;
#if REPLAY_TIMING
	replay_timing_init();
#endif
#if PREFETCH
	prefetch_init();
#endif

	//Create initial kthread_data object and run thread_execute()
	kthread_data = KMALLOC(sizeof(kthread_data_t),GFP_KERNEL);
	if(!kthread_data) {
	  print_error_nomem();
	  err = -ENOMEM;
	  goto out;
	}
	sem = KMALLOC(sizeof(struct semaphore),GFP_KERNEL);
	if(!sem) {
	  print_error_nomem();
	  err = -ENOMEM;
	  goto out;
	}
	kthread_data->sem = sem;
	suspend = KMALLOC(sizeof(atomic_t),GFP_KERNEL);
	if(!suspend) {
	  print_error_nomem();
	  err = -ENOMEM;
	  goto out;
	}
	kthread_data->suspend = suspend;
	atomic_set(kthread_data->suspend,0);
	sema_init(kthread_data->sem,1);
	down(kthread_data->sem);
	kthread_data->srcmdp = &r->srcmdp[r->cur_cmd];
	kthread_data->r = r;
	kthread_data->terminate = 0;
	kthread_data->task = current;
	kthread_data->thread_loc = 0;
	kthread_data->initialized = 1;

	//Execute
	err = thread_execute(kthread_data);

	KFREE(kthread_data->sem);
	KFREE(kthread_data->suspend);
	KFREE(kthread_data);

out:
	print_debug_message("# of replay commands = %d\nLast Executed Command # = %d\nLast Executed Command = %d\nerr = %d",r->cmd_num,r->cur_cmd,cmd,err);

#if PREFETCH
	prefetch_destroy();
#endif
#if REPLAY_TIMING
	replay_timing_release();
#endif
	print_exit_status(err);
	return err;
}

#else //#if MULTITHREADING
int do_replay(struct replay* r)
{
	int err = 0;
	int cmd = -1, relcmd=0;
	struct replay_command* scmdp;

	print_entry_location();
	r->cur_cmd = 0;
#if PREFETCH
	prefetch_init();
#endif
	replay_timing_init();

	while ((scmdp = fetch_cmd(r)) ) {
	        cmd = scmdp->command;
#if PREFETCH
		prefetch_readahead(r,r->data);
#endif
#if REPLAY_TIMING
		replay_timing_pre_op(scmdp,REPLAY_TIMING_ELAPSED,REPLAY_TIMING_SPEED);
#endif
		//To keep a count of the cmd number actually inserted by tracefs that is being currently played.
		// only for debugging purposes.
		if(scmdp->command == OP_LOOKUP || scmdp->command == OP_OPEN || scmdp->command == OP_CREATE ||
                         scmdp->command == OP_FLUSH || scmdp->command == OP_RELEASE || scmdp->command == OP_SYMLINK ||
                         scmdp->command == OP_READPAGE || scmdp->command == OP_COMMIT_WRITE || scmdp->command == OP_LINK ||
                         scmdp->command == OP_WRITEPAGE || scmdp->command == OP_MKDIR || scmdp->command == OP_RMDIR ||
                         scmdp->command == OP_UNLINK || scmdp->command == OP_D_RELEASE || scmdp->command == OP_READDIR ||
                         scmdp->command == OP_RENAME || scmdp->command == OP_MKNOD || scmdp->command == OP_SETATTR ||
                         scmdp->command == OP_FSYNC || scmdp->command == OP_SYNC_PAGE || scmdp->command == OP_LLSEEK ||
                         scmdp->command == OP_PERMISSION || scmdp->command == OP_READ)
		{
			++relcmd;
  			 //printk("\n# %d CMD %d ",relcmd,scmdp->command);
		}
		switch (scmdp->command) {
#if PREFETCH
			case OP_LOAD:		REPLAY_CMD(replay_prefetch_load);
						break;
			case OP_UNLOAD:		REPLAY_CMD(replay_prefetch_unload);
						break;
#else
			case OP_LOAD:		REPLAY_CMD(replay_load);
						break;
			case OP_UNLOAD:		REPLAY_CMD(replay_unload);
						break;
#endif
			case OP_CREATE:		REPLAY_CMD(replay_create);
						break;
			case OP_LINK:		REPLAY_CMD(replay_link);
						break;
			case OP_SYMLINK:	REPLAY_CMD(replay_symlink);
						break;
			case OP_RENAME:	        REPLAY_CMD(replay_rename);
						break;
			case OP_OPEN:		REPLAY_CMD(replay_open);
						break;
			case OP_LOOKUP:		REPLAY_CMD(replay_lookup);
						break;
		        case OP_FLUSH:		REPLAY_CMD(replay_flush);
						break;
			case OP_RELEASE:	REPLAY_CMD(replay_release);
						break;
			case OP_WRITE:		REPLAY_CMD(replay_write);
						break;
			case OP_READ:		REPLAY_CMD(replay_read);
						break;
			case OP_MKDIR:		REPLAY_CMD(replay_mkdir);
						break;
			case OP_READPAGE:	REPLAY_CMD(replay_readpage);
						break;
			case OP_READPAGES:	REPLAY_CMD(replay_readpages);
                        			break;
			case OP_RMDIR:          REPLAY_CMD(replay_rmdir);
						break;
			case OP_COMMIT_WRITE:	REPLAY_CMD(replay_commit_write);
						break;
			/*The sync page operation tries to call find_get_page to fetch a
			page from the mapping at a particular offset. During replay it may
			so happen that the pdflush kernel thread has already woken up and
			flushes the corresponding dirty pages of this mapping and this page is
			no longer dirty. As a result the function find_get_page returns a NULL
			page. This can happen due to the difference in system load at the time
			of tracing and replaying. Hence the syncpage commands are not replayed
			at this moment.*/
			case OP_SYNC_PAGE:		REPLAY_CMD(replay_sync_page);
                    				break;
			case OP_UNLINK:		REPLAY_CMD(replay_unlink);
						break;
			case OP_D_RELEASE:      REPLAY_CMD(replay_d_release);
						break;
			case OP_READDIR:	REPLAY_CMD(replay_readdir);
						break;
        		case OP_WRITEPAGE:      REPLAY_CMD(replay_writepage);
		                                break;
            		case OP_DGET:           REPLAY_CMD(replay_dget);
		                                break;
		        case OP_DPUT:           REPLAY_CMD(replay_dput);
		                                break;
        		case OP_MKNOD:          REPLAY_CMD(replay_mknod);
		                                break;
        		case OP_SETATTR:        REPLAY_CMD(replay_setattr);
		                                break;
        		case OP_FSYNC:          REPLAY_CMD(replay_fsync);
		                                break;
        		case OP_LLSEEK:         REPLAY_CMD(replay_llseek);
		                                break;
        		case OP_PERMISSION:     REPLAY_CMD(replay_permission);
		                                break;
			default:
				printk(KERN_WARNING"Unknown replay command encoutered, command = %d - aborting\n",scmdp->command);
				err = -EINVAL;
				goto out;
		}

 		//Output '.' every 1000 commands
 		if(!(r->cur_cmd % 1000))
 		  printk(".");
/* #if REPLAY_TIMING
 		replay_timing_post_op();
 #endif */
	}
out:
	printk("# of replay commands = %d\nLast Executed Command # = %d\nLast Executed Command = %d\nerr = %d",r->cmd_num,r->cur_cmd,cmd,err);

#if PREFETCH
	prefetch_destroy();
#endif
	print_exit_status(err);
	return err;
}
#endif //#if MULTITHREADING

int parse_replayfile(char* replay_fil, struct replay* r)
{
	int error = 0;
	char* cpkTraceName = NULL;
	char* cpkDataName = NULL;
	char* cpk = NULL;
	mm_segment_t fs_old;
	print_entry_location();

	r->trace = NULL;
	r->data = NULL;

	if (!replay_fil) {// || strlen_user(replay_fil) < 2) {
		printk(KERN_WARNING"No replay file specified\n");
		error = -EINVAL;
		goto out;
	}

	/*cpkTraceName = getname(replay_fil);
	if (IS_ERR(cpkTraceName)) {
		printk(KERN_WARNING "Failed to getname\n");
		error = PTR_ERR(cpkTraceName);
		cpkTraceName = NULL;
		goto out;
	}*/
	cpkTraceName = replay_fil;

	error = strlen(cpkTraceName); // should be strlen_user();
	if(error <= 0 || error >= PATH_MAX) {
		error = -EINVAL;
		goto out;
	}

	cpkDataName = KMALLOC(PATH_MAX, GFP_KERNEL);
	if (!cpkDataName) {
		error = -ENOMEM;
		goto out;
	}

	strcpy(cpkDataName, cpkTraceName);
	strcat(cpkDataName, ".data");

	r->data = filp_open(cpkDataName, O_RDONLY, 0);
	if (!r->data || IS_ERR(r->data)) {
		error = PTR_ERR(r->data);
		r->data = NULL;
		goto out;
	}

	cpk = KMALLOC(PAGE_SIZE, GFP_KERNEL);
	if (!cpk) {
		error = -ENOMEM;
		goto out;
	}

	r->trace = filp_open(cpkTraceName, O_RDONLY, 0);
	if (!r->trace || IS_ERR(r->trace)) {
		error = PTR_ERR(r->trace);
		r->trace = NULL;
		goto out;
	}

	/*printk(KERN_WARNING "TODO: check if regular file\n");*/

	if (!r->trace->f_op || !r->trace->f_op->read) {
		error = -EINVAL;
		goto out;
	}

	fs_old = get_fs(); set_fs(KERNEL_DS);

  /*Trace file - 1. Trace head + 2. Trace RAT*/
	error = r->trace->f_op->read(r->trace, cpk,
	    sizeof(struct trace_head) + sizeof(struct trace_rat), &r->trace->f_pos);

	set_fs(fs_old);

	if (error < 0)
		goto out;
	/*if (error < sizeof(struct trace_head) + sizeof(struct trace_rat) ||
	    ((int)cpk) != (int)MAGIC_TRACE ||*/
	if (((struct trace_head*)cpk)->version != TRACE_VER) {
		printk(KERN_WARNING "Wrong file version\n");
		error = -EINVAL;
		goto out;
	}

	/* RAT */
  /*Effectively (struct trace_rat*)(cpk+sizeof(struct trace_head))->num */
	r->res_num = ((int*)cpk)[3];

	//Using vmalloc, kmalloc insufficient for large traces
	r->ratp = (struct res_alloc_table*)VMALLOC(r->res_num * sizeof(struct res_alloc_table));
	if (!r->ratp) {
		error = -ENOMEM;
		goto out;
	}

	fs_old = get_fs(); set_fs(KERNEL_DS);
	error = r->trace->f_op->read(r->trace, (char*)r->ratp,
	    r->res_num * sizeof(struct res_alloc_table), &r->trace->f_pos);

	set_fs(fs_old);

	if (error < 0)
		goto out;

	/* CMDS */
	fs_old = get_fs(); set_fs(KERNEL_DS);
	error = r->trace->f_op->read(r->trace, cpk, sizeof(struct trace_cmd), &r->trace->f_pos);
	set_fs(fs_old);
	if (error < 0)
		goto out;

	r->cmd_num = ((int*)cpk)[1];

	//Using vmalloc, kmalloc insufficient for large traces
	//Allocate space for the PROG
	r->srcmdp = (struct replay_command*)VMALLOC(r->cmd_num * sizeof(struct replay_command));
	if (!r->srcmdp) {
		error = -ENOMEM;
		goto out;
	}

	fs_old = get_fs(); set_fs(KERNEL_DS);
	error = r->trace->f_op->read(r->trace, (char*)r->srcmdp,
	    r->cmd_num * sizeof(struct replay_command), &r->trace->f_pos);

	set_fs(fs_old);
	if (error < 0)
		goto out;

out:
	if (cpk)
	  KFREE(cpk);
	if (cpkDataName)
	  KFREE(cpkDataName);
	if (r->trace)
	  filp_close(r->trace, NULL);
	/*if (r->data)
		filp_close(r->data, NULL);
	if(cpkTraceName)
		putname(cpkTraceName);*/
	print_exit_status(error);
	return error;
}

void cleanup_rat(struct replay* r)
{
	int i, dr;

	print_entry_location();

	if (!r->ratp)
	  goto out;
	for(i = MIN_RAT_SIZE; i < r->res_num; i++) {
	  /* fput all open files */
	  if (r->ratp[i].type == ARG_FILE) {
	    if (r->ratp[i].count > 0 && r->ratp[i].u.file) {
	      if (r->ratp[i].u.file->f_op && r->ratp[i].u.file->f_op->flush) {
		lock_kernel();
		r->ratp[i].u.file->f_op->flush( r->ratp[i].u.file );
		unlock_kernel();
	      }
	      while(r->ratp[i].count > 0) {
		/* reduce ref count of corresponding dentry */
		dr = lookup_rat(r, (int)r->ratp[i].u.file->f_dentry, ARG_DENTRY);
		if (dr > 0 && r->ratp[dr].count > 0) {
		  printk(KERN_INFO"dput: %p\n", r->ratp[dr].u.dentry);
		  r->ratp[i].count--;
		}
		/* put the file itself */
		printk(KERN_INFO"Free file: %p\n", r->ratp[i].u.file);
		fput(r->ratp[i].u.file);
		r->ratp[i].count--;
	      }
	      r->ratp[i].u.file = NULL;
	    }
	    r->ratp[i].type = ARG_INVAL;
	  }

	  /* dput all dentries we dgot :-) */
	  if (r->ratp[i].type == ARG_DENTRY) {
	    if (r->ratp[i].count > 0 && r->ratp[i].u.dentry) {
	      while(r->ratp[i].count > 0) {
		fist_print_dentry("cleanup", r->ratp[i].u.dentry);
		fist_print_dentry("cleanupp", r->ratp[i].u.dentry->d_parent);
		fist_print_dentry("cleanuppp", r->ratp[i].u.dentry->d_parent->d_parent);
		dput(r->ratp[i].u.dentry);
		r->ratp[i].count--;
	      }
	      r->ratp[i].u.dentry = NULL;
	    }
	    r->ratp[i].type = ARG_INVAL;
	  }
	  /* easy: free memory buffers */
	  if (r->ratp[i].type == ARG_STR) {
	    r->ratp[i].type = ARG_INVAL;
	    if(r->ratp[i].u.cpval != NULL)
	      VFREE(r->ratp[i].u.cpval);
	    r->ratp[i].u.cpval = NULL;
	  }
#if MULTITHREADING
	  /* stop and deallocate threads and semaphores */
	  if(r->ratp[i].type == ARG_KTHREAD_DATA) {
	    if(r->ratp[i].u.kthread_data) {
	      if(r->ratp[i].u.kthread_data->sem) {
		r->ratp[i].u.kthread_data->terminate = 1;
		down(r->ratp[i].u.kthread_data->sem); //Make sure it's done
		r->ratp[i].u.kthread_data->task->state = TASK_RUNNING;
		if(init_thread_loc != r->ratp[i].u.kthread_data->thread_loc)
		  kthread_stop(r->ratp[i].u.kthread_data->task);
		KFREE(r->ratp[i].u.kthread_data->sem);
	      }
	      KFREE(r->ratp[i].u.kthread_data);
	    }
	  }
#endif
	}

	/* print unallocated resources if any */
	printk(KERN_INFO"Unfreed resources:\n");
	for(i = MIN_RAT_SIZE; i < r->res_num; i++) {
	  if (r->ratp[i].type != ARG_INVAL &&
	      r->ratp[i].type != ARG_INT_FLAGS)
	    printk(KERN_INFO"   RAT[%u] = {%u, 0x%x, %i}\n", i, r->ratp[i].type, r->ratp[i].u.ival, r->ratp[i].count);
	}
 out:
	print_exit_status(0);
}

int lookup_rat(struct replay* r, int ival, int type)
{
	int i;

	for (i = MIN_RAT_SIZE; i< r->res_num; i++) {
		if(r->ratp[i].type == type && r->ratp[i].u.ival == ival)
			return i;
	}
	return -1;
}

#ifdef FIST_MALLOC_DEBUG
/* for malloc debugging */
static atomic_t replayfs_kmalloc_counter;
static atomic_t replayfs_kmalloc_counter_diff;
static atomic_t replayfs_vmalloc_counter;
static atomic_t replayfs_vmalloc_counter_diff;

void *
replayfs_kmalloc(size_t len, int flag, int line, const char *file)
{
    void *ptr = (void *) kmalloc(len, flag);
    if (ptr) {
	atomic_inc(&replayfs_kmalloc_counter);
	atomic_inc(&replayfs_kmalloc_counter_diff);
	//printk(KERN_ALERT "KM:%d:%p:%d:%s\n", atomic_read(&replayfs_kmalloc_counter),ptr, line, file);
	//printk(KERN_ALERT "KM:%d:%d:%p:%d:%s\n", atomic_read(&replayfs_kmalloc_counter),atomic_read(&replayfs_kmalloc_counter_diff),ptr, line, file);
    }
    return ptr;
}

void
replayfs_kfree(void *ptr, int line, const char *file)
{
    atomic_inc(&replayfs_kmalloc_counter);
    atomic_dec(&replayfs_kmalloc_counter_diff);
    //printk(KERN_ALERT "KF:%d:%p:%d:%s\n", atomic_read(&replayfs_kmalloc_counter),ptr, line, file);
    //printk(KERN_ALERT "KF:%d:%d:%p:%d:%s\n", atomic_read(&replayfs_kmalloc_counter),atomic_read(&replayfs_kmalloc_counter_diff),ptr, line, file);
    kfree(ptr);
}

void*
replayfs_vmalloc(size_t len,int line, const char *file)
{
    void *ptr = (void *) vmalloc(len);
    if (ptr) {
	atomic_inc(&replayfs_vmalloc_counter);
	atomic_inc(&replayfs_vmalloc_counter_diff);
	printk(KERN_ALERT "KM:%d:%p:%d:%s\n", atomic_read(&replayfs_vmalloc_counter),ptr, line, file);
	//printk(KERN_ALERT "VM:%d:%d:%p:%d:%s\n", atomic_read(&replayfs_vmalloc_counter),atomic_read(&replayfs_vmalloc_counter_diff),ptr, line, file);
    }
    return ptr;
}

void
replayfs_vfree(void *ptr, int line, const char *file)
{
    atomic_inc(&replayfs_vmalloc_counter);
    atomic_dec(&replayfs_vmalloc_counter_diff);
    printk(KERN_ALERT "KF:%d:%p:%d:%s\n", atomic_read(&replayfs_vmalloc_counter),ptr, line, file);
    //printk(KERN_ALERT "VF:%d:%d:%p:%d:%s\n", atomic_read(&replayfs_vmalloc_counter),atomic_read(&replayfs_vmalloc_counter_diff),ptr, line, file);
    vfree(ptr);
}
#endif /* FIST_MALLOC_DEBUG */


static char* replay_mntpoint = NULL;
static char* replay_file = NULL;
static int b_compare_data;

int sys_replay(void)
{
	struct nameidata nd;
	int error;
	struct replay r;
	r.ratp = NULL;

#ifdef FIST_MALLOC_DEBUG
	atomic_set(&replayfs_kmalloc_counter,0);
	atomic_set(&replayfs_kmalloc_counter_diff,0);
	atomic_set(&replayfs_vmalloc_counter,0);
	atomic_set(&replayfs_vmalloc_counter_diff,0);
#endif //#ifdef FIST_MALLOC_DEBUG

	printk("replay_file = %s\n", replay_file);
	error = parse_replayfile(replay_file, &r);
	if (error < 0)
		goto out;

	r.ratp[RAT_ROOT_MNT].u.mnt   = NULL;
	r.ratp[RAT_ROOT_DENTRY].u.dentry= NULL;

	/* absolute or relative to the caller's pwd */
	if(!replay_mntpoint) {
		printk(KERN_WARNING"No mountpoint specified\n");
		error = -EINVAL;
		goto out;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (path_init(replay_mntpoint, LOOKUP_FOLLOW, &nd))
		error = path_walk(replay_mntpoint, &nd);
#else
	error = path_lookup(replay_mntpoint,LOOKUP_FOLLOW,&nd);
#endif
	if (error) {
		printk(KERN_WARNING"Wrong mountpoint\n");
		error = -EINVAL;
		goto out;
	}

	if(b_compare_data < 1)
		b_compare_data = 0;
	else
		b_compare_data = 1;
	r.b_compare_data = b_compare_data;
	r.ratp[RAT_ROOT_MNT].type  = ARG_VFSMNT;
	r.ratp[RAT_ROOT_MNT].u.mnt   = nd.mnt;
	r.ratp[RAT_ROOT_MNT].count = 1;
	r.ratp[RAT_ROOT_DENTRY].type  = ARG_DENTRY;
	r.ratp[RAT_ROOT_DENTRY].u.dentry= nd.dentry;
	r.ratp[RAT_ROOT_DENTRY].count = 1;
	error = do_replay(&r);
out:
	if(error)
	  printk(KERN_ALERT "\nFAILED!!!\n# of cmd attempted: %d, Last cmd: %d err = %d\n",r.cmd_num,(r.srcmdp)->command,error);
	else
	  printk(KERN_ALERT "Replay Successful!\n");

	printk("Cleaning up RAT\n");
	cleanup_rat(&r);
	printk("RAT: Cleaned!\n");

	if(r.data)
		filp_close(r.data, NULL);
	if(r.ratp && r.ratp[RAT_ROOT_DENTRY].u.dentry)
		dput(r.ratp[RAT_ROOT_DENTRY].u.dentry);
	if(r.ratp && r.ratp[RAT_ROOT_MNT].u.mnt) {
	  r.ratp[RAT_ROOT_MNT].count--; //Tim: DEBUG
		while (r.ratp[RAT_ROOT_MNT].count--) {
			mntput(r.ratp[RAT_ROOT_MNT].u.mnt);
		}
	}
	if(r.srcmdp)
	  VFREE(r.srcmdp);
	if(r.ratp)
	  VFREE(r.ratp);

	printk("Done!\n");
	return error;
}
//Added init_replayfs and exit_replayfs to be inline with 2.6 modules
static int __init init_replayfs(void)
{
    printk("Registering replayfs version $Id: replay.c,v 1.86 2007/01/30 07:07:03 gopalas Exp $\n");
    return sys_replay();
}
static void __exit exit_replayfs(void)
{
}

//EXPORT_NO_SYMBOLS;
MODULE_AUTHOR("Nikolai Joukov & Akshat Aranya & Timothy T. Wong");
MODULE_DESCRIPTION("TraceFS replay engine");
MODULE_LICENSE("GPL");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
MODULE_PARM(replay_mntpoint, "s");
MODULE_PARM_DESC(replay_mntpoint, "Lower level mount point");

MODULE_PARM(replay_file, "s");
MODULE_PARM_DESC(replay_file, "File to replay");

MODULE_PARM(b_compare_data, "i");
MODULE_PARM_DESC(b_compare_data, "Compare the data or not");


MODULE_PARM(fist_debug_var, "i");
MODULE_PARM_DESC(fist_debug_var, "Debug level setting");

#else
module_param(replay_mntpoint, charp, 0644);
MODULE_PARM_DESC(replay_mntpoint, "Lower level mount point");

module_param(replay_file, charp, 0644);
MODULE_PARM_DESC(replay_file, "File to replay");

module_param(b_compare_data, int, 0644);
MODULE_PARM_DESC(b_compare_data, "Compare the data or not");


module_param(fist_debug_var, int, 0644);
MODULE_PARM_DESC(fist_debug_var, "Debug level setting");
#endif
module_init(init_replayfs)
module_exit(exit_replayfs)

