/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#define NEED_DESCRIPTIVE_MESSAGES
#define _DECLARE_SYSCALL_TABLE_
#include "fist.h"
#include "common.h"
#include "ad_stream.h"
#include "stream_header.h"
#include "serializer.h"
#include <linux/utsname.h>
#include "md5.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif

#define MD5_CKSUM_LEN 16
struct stream_pd {
    char *process_bmap;
    pid_t max_process_seen;
    unsigned int verbosity;
};

#define pd(ptr) (((struct stream_pd *)(ptr)->ad_private_data))
#define VERBOSITY(this, bit) (pd((this))->verbosity & (bit))

#define ALLOC_CONTEXT *context = (void*)KMALLOC(sizeof(msg_seqnum_t),GFP_KERNEL);
#define ALLOC_CONTEXT_NOFS do { \
  *context = (void*)KMALLOC(sizeof(msg_seqnum_t),GFP_NOFS); \
  if (!(*context)) { \
    down(&reserved_context_sem); \
    *context = (void*)&reserved_context; \
  } \
} while(0);

#define SET_CONTEXT **((int**)context) = atomic_read(&sequence_number); \
                    atomic_inc(&sequence_number);
#define WRITE_CONTEXT memcpy(buf->data+SEQNUM_LOC,*context,sizeof(msg_seqnum_t));
#define FREE_CONTEXT do{if(context)KFREE(*context); }while(0);

#define FREE_CONTEXT_NOFS do { \
  if(*context != &reserved_context) {\
    if(*context) KFREE(*context); \
  }else \
    up(&reserved_context_sem); \
}while(0);

/*
 * This is primarily for writepages, when memory is low, writepage() is called
 * to write to disc; if memory allocation is attempted during writepage() when
 * memory is low, tracefs will enter into an infinite loop, these reserved
 * variables are used to prevent this from happening.
 */
static msg_seqnum_t reserved_context; //Reserved for writepage if ENOMEM
static struct semaphore reserved_context_sem; //Semaphore for the reserve_context
static struct buffer reserved_buffer; //Reserved for writepages if ENOMEM
static struct semaphore reserved_buffer_sem; //Semphore for the reserved_buffer

// determine if we've seen this process before
static inline int get_pbit(struct assembly_driver *this, pid_t pid)
{
    ASSERT(pid >= 0 && pid < 0xFFFF);
    return (*(pd(this)->process_bmap + (pid >> 3)) & (0x1 << (pid & 0x7)));
}

// yes we saw this process
static inline void set_pbit(struct assembly_driver *this, pid_t pid)
{
    *(pd(this)->process_bmap + (pid >> 3)) |= (0x1 << (pid & 0x7));
}


static inline void add_timestamp(struct assembly_driver *this, struct buffer *buf)
{
    struct timeval now;
    if (VERBOSITY(this, STR_TIMESTAMP))
    {
	    do_gettimeofday(&now);
    	if (VERBOSITY(this, STR_UTIMESTAMP))
      {
		    add_tv_ptr(buf, ARG_UTIMESTAMP, sizeof(now), &now);
	    }
      else
      {
    		add_tv(buf, ARG_TIMESTAMP, now.tv_sec);
      }
    }
}

#define start_pre_trace_message(this, dest, buf, op) add_op(this, dest, buf, pre_op((op)))
#define start_post_trace_message(this, dest, buf, op) add_op(this, dest, buf, post_op((op)))

#define PRE_HEADER(op) \
        int header_seqnum;  \
        void* header_buf; \
	struct buffer *buf; \
        buf = get_buffer(); \
	if (!buf) {\
                *context = NULL; \
		goto out; \
        } \
	start_pre_trace_message(this, dest, buf, (op)); \
	add_timestamp(this, buf); \
        ALLOC_CONTEXT; \
        SET_CONTEXT; \
        WRITE_CONTEXT; \
        header_buf = &header_seqnum; \
        memcpy(header_buf,buf->data+SEQNUM_LOC,sizeof(msg_seqnum_t)); \
	if (VERBOSITY(this, STR_ARGS)) {

#define PRE_HEADER_NOFS(op) \
        int header_seqnum;  \
        void* header_buf; \
	struct buffer *buf; \
        buf = get_buffer_nofs(); \
	if (!buf) { \
	     down(&reserved_buffer_sem); \
             buf = &reserved_buffer;  \
             buf->ptr = buf->data + INIT_LOC; \
	} \
	start_pre_trace_message(this, dest, buf, (op)); \
	add_timestamp(this, buf); \
        ALLOC_CONTEXT_NOFS; \
        SET_CONTEXT; \
        WRITE_CONTEXT; \
        header_buf = &header_seqnum; \
        memcpy(header_buf,buf->data+SEQNUM_LOC,sizeof(msg_seqnum_t)); \
	if (VERBOSITY(this, STR_ARGS)) {

#define PRE_HEADER_NO_POST(op) \
        int header_seqnum;  \
        void* header_buf; \
	struct buffer *buf; \
        buf = get_buffer(); \
	if (!buf) \
		goto out; \
	start_pre_trace_message(this, dest, buf, (op)); \
	add_timestamp(this, buf); \
        ALLOC_CONTEXT; \
        SET_CONTEXT; \
        WRITE_CONTEXT; \
        FREE_CONTEXT; \
        header_buf = &header_seqnum; \
        memcpy(header_buf,buf->data+SEQNUM_LOC,sizeof(msg_seqnum_t)); \
	if (VERBOSITY(this, STR_ARGS)) {

#define POST_HEADER(op)\
        struct buffer *buf; \
        int header_seqnum; \
        void* header_buf; \
        buf = get_buffer(); \
	if (!buf) {\
          FREE_CONTEXT; \
	  goto out; \
        } \
        WRITE_CONTEXT; \
        FREE_CONTEXT; \
        header_buf = &header_seqnum; \
        memcpy(header_buf,buf->data+SEQNUM_LOC,sizeof(msg_seqnum_t)); \
	if (VERBOSITY(this, STR_POST_OP)) { \
		start_post_trace_message(this, dest, buf, (op)); \
		add_timestamp(this, buf); \
		if (VERBOSITY(this, STR_ARGS)) {

#define POST_HEADER_NOFS(op)\
        struct buffer *buf; \
        int header_seqnum; \
        void* header_buf; \
        buf = get_buffer_nofs(); \
	if (!buf) { \
            down(&reserved_buffer_sem); \
            buf = &reserved_buffer; \
            buf->ptr = buf->data + INIT_LOC; \
        } \
        WRITE_CONTEXT; \
        FREE_CONTEXT_NOFS; \
        header_buf = &header_seqnum; \
        memcpy(header_buf,buf->data+SEQNUM_LOC,sizeof(msg_seqnum_t)); \
	if (VERBOSITY(this, STR_POST_OP)) { \
		start_post_trace_message(this, dest, buf, (op)); \
		add_timestamp(this, buf); \
		if (VERBOSITY(this, STR_ARGS)) {

#define PRE_FOOTER \
	} \
	end_trace_message(buf); \
	down(&this->ad_odriver_sem); \
	if (dest->td_state != TD_ERROR) { \
		if (dest->td_od->od_ops->write(dest->td_od, buf->data, buf->ptr - buf->data,OO_KERNEL_BUFFER) != buf->ptr - buf->data) \
			dest->td_state = TD_ERROR; \
	} \
	up(&this->ad_odriver_sem); \
        if(buf != &reserved_buffer) \
          release_buffer(buf); \
        else { \
          up(&reserved_buffer_sem);  \
	}\
	out: \
	    return;

#define POST_FOOTER \
		} \
		end_trace_message(buf); \
		down(&this->ad_odriver_sem); \
		if (dest->td_state != TD_ERROR) { \
			if (dest->td_od->od_ops->write(dest->td_od, buf->data, buf->ptr - buf->data,OO_KERNEL_BUFFER) != buf->ptr - buf->data) \
				dest->td_state = TD_ERROR; \
		} \
		up(&this->ad_odriver_sem); \
                if(buf != &reserved_buffer) \
            	  release_buffer(buf); \
                else \
                  up(&reserved_buffer_sem); \
	} \
	out: \
            return;


// adds a process name message to the trace
static void add_pname_message(struct assembly_driver *this, struct trace_destination *dest)
{
    /*
     * NOTE! NOTE!! fix the value of available in case this log message is ever changed
     * --------------------------------------------------------------------------------------------------------------
     * |              |              |              |         |     |           |        |                          |
     * |  MSG_PNAME   |  MSG_LENGTH  |  MSG_SEQNUM  | ARG_PID | pid | ARG_PNAME | length | <------ available -----> |
     * |              |              |              |         |     |           |        |                          |
     * --------------------------------------------------------------------------------------------------------------
     * <-------------------------------------- DATA_SIZE ----------------------------------------------------------->
     * <--------- INIT_LOC ------------------------->
     */

#define PARGS_AVAILABLE  (DATA_SIZE - INIT_LOC - sizeof(field_t) - sizeof(pid_t) - sizeof(field_t) - sizeof(field_len_t))

    struct buffer *buf = get_buffer();
    unsigned long length;
    msg_seqnum_t seqnum;
    int err;
    ASSERT(dest);
    ASSERT(dest->td_od);
    if (buf) {
	op_t tmp = MSG_PNAME;
	memcpy(buf->data, &tmp, sizeof(tmp));



	//Add Sequence Number
	seqnum = atomic_read(&sequence_number);
	atomic_inc(&sequence_number);
	memcpy(buf->data+SEQNUM_LOC,&seqnum,sizeof(seqnum));

	add_tv(buf, ARG_PID, current->pid);
	length = current->mm->arg_end - current->mm->arg_start;
	if (length > PARGS_AVAILABLE)
	    length = PARGS_AVAILABLE;
	if (length) {
	    field_t tmp = ARG_PNAME;
	    field_len_t len = length;
	    tmp |= FIELD_LENGTH_MASK;
	    memcpy((buf)->ptr, &tmp, sizeof(field_t));
	    (buf)->ptr += sizeof(field_t);
	    memcpy((buf)->ptr, &len, sizeof(len));
	    (buf)->ptr += sizeof(field_len_t);
	    err = copy_from_user(buf->ptr, (char *)current->mm->arg_start, len);
	    (buf)->ptr += len;
	}
	end_trace_message(buf);
	down(&this->ad_odriver_sem);
	if (dest->td_state != TD_ERROR) {
		if (dest->td_od->od_ops->write(dest->td_od, buf->data, buf->ptr - buf->data,OO_KERNEL_BUFFER) != buf->ptr - buf->data)
			dest->td_state = TD_ERROR;
	}
	up(&this->ad_odriver_sem);
	release_buffer(buf);
    }
}
#ifdef TRACE_SYSCALL
#include "syscalls.h"
static void add_syscall_messages(struct assembly_driver *this, struct trace_destination *dest)
{
	struct buffer *buf = get_buffer();
	int i;
        op_t op = MSG_SYSCALL;
	msg_seqnum_t seqnum;
	if (!buf) goto out;

	for (i = 0; __syscalls[i].name != NULL; i++) {
                //Add op
         	memcpy(buf->data, &op, sizeof(op));

		//Add Sequence Number
		seqnum = atomic_read(&sequence_number);
		atomic_inc(&sequence_number);
		memcpy(buf->data+SEQNUM_LOC,&seqnum,sizeof(seqnum));

		add_tv(buf, ARG_SYSCALL, __syscalls[i].id);
		add_tlv(buf, ARG_SYSCALL_NAME, strlen(__syscalls[i].name), __syscalls[i].name);
		end_trace_message(buf);
		down(&this->ad_odriver_sem);
		if (dest->td_state != TD_ERROR) {
			if (dest->td_od->od_ops->write(dest->td_od, buf->data, buf->ptr - buf->data) != buf->ptr - buf->data)
				dest->td_state = TD_ERROR;
		}
		up(&this->ad_odriver_sem);
		reset_buffer(buf);
	}
	release_buffer(buf);
out:
	return;
}
#else
#define add_syscall_messages(a, b)
#endif

static void add_op(struct assembly_driver *this,
          struct trace_destination *dest, struct buffer *buf, op_t op)
{
	op_t tmp = op;

	memcpy(buf->data, &tmp, sizeof(tmp));

	if (VERBOSITY(this, STR_PID)) add_tv(buf, ARG_PID, current->pid);
#ifdef TRACE_SYSCALL
	if (VERBOSITY(this, STR_SYSCALL_ID)) add_tv(buf, ARG_SYSCALL, current->sys_call);
#endif
	if (VERBOSITY(this, STR_UID)) add_tv(buf, ARG_UID, current->euid);
	if (VERBOSITY(this, STR_GID)) add_tv(buf, ARG_GID, current->egid);
	if (VERBOSITY(this, STR_PNAME) && current->mm) {
		if (!get_pbit(this, current->pid)) {
			// If this process is not already seen
			set_pbit(this, current->pid);
			add_pname_message(this, dest);
			if (current->pid > pd(this)->max_process_seen)
				pd(this)->max_process_seen = current->pid;
		} else if (current->pid < pd(this)->max_process_seen) {
			// If we've seen this process before, we're assuming a roll over
			clear_page(pd(this)->process_bmap);
			set_pbit(this, current->pid);
			pd(this)->max_process_seen = current->pid;
			add_pname_message(this, dest);
		}
	}

}


static void add_inode(struct buffer *buf, struct inode *inode)
{
    add_tv(buf, ARG_INODE_PTR, inode);
    /* for now, just add the inum */
    add_tv(buf, ARG_INUM, inode->i_ino);
}

static inline void add_dentry(struct buffer *buf, struct dentry *dentry)
{
    add_tv(buf, ARG_DENTRY_PTR, dentry);
    /* for now, just add the name */
    add_tlv(buf, ARG_PATHNAME, dentry->d_name.len, dentry->d_name.name);
    //Add Inode Number - so we can hash the dentry on the inode later
	//The below add was changed from dentry->d_inode to dentry->d_inode->i_ino since
	//it is the inode number we are adding and not the inode pointer.
    if(dentry->d_inode)
      add_tv(buf,ARG_INUM,dentry->d_inode->i_ino);
}

static void add_file(struct buffer *buf, file_t *file)
{
    add_tv(buf, ARG_FILE_PTR, file);
    add_tv(buf, ARG_FILE_FLAGS, file->f_flags);
    /* for now, just add the inum  and dentry */
    add_tv(buf, ARG_INUM, file->f_dentry->d_inode->i_ino);
    add_tlv(buf, ARG_PATHNAME, file->f_dentry->d_name.len, file->f_dentry->d_name.name);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static void add_pages(struct buffer *buf,struct list_head* pages)
{
  struct list_head* list;
  struct page* page;
  struct page* first_page;

  first_page = NULL;

  //Assume that the list is circular
  list_for_each(list,pages) {
    page = list_entry(list,struct page,lru);
    if(first_page == page)
      break;
    else if(first_page == NULL)
      first_page = page;
    add_tv(buf,ARG_PAGES_INDEX,page->index);
  }
}

static void add_pages_with_data(struct buffer *buf,struct list_head* pages)
{
  struct list_head *list;
  struct page* page;
  struct page* first_page;

  first_page = NULL;

  //Assume that the list is circular
  list_for_each(list,pages) {
    page = list_entry(pages,page_t,lru);
    if(first_page == page)
      break;
    else if(first_page == NULL)
      first_page = page;
    add_tv(buf,ARG_PAGES_INDEX,page->index);
  }
}
#endif

static void add_page(struct buffer *buf, struct page* page)
{
	add_tv(buf, ARG_PAGE_INDEX, page->index);
}

static void add_page_with_data(struct buffer *buf, struct page* page)
{
	add_tv(buf, ARG_PAGE_INDEX, page->index);
}

/* used by functions that have two dentry objects */
static inline void add_dentry2(struct buffer *buf, struct dentry *dentry)
{
    /* for now, just add the name */
    add_tv(buf, ARG_NEW_DENTRY_PTR, dentry);
    add_tlv(buf, ARG_NEW_PATHNAME, dentry->d_name.len, dentry->d_name.name);
}

/* used by functions that have two inode objects */
static void add_inode2(struct buffer *buf, struct inode *inode)
{
    add_tv(buf, ARG_NEW_INODE_PTR, inode);
    /* for now, just add the inum */
    add_tv(buf, ARG_NEW_INUM, inode->i_ino);
}

static inline void add_flags(struct buffer *buf, int flags)
{
    add_tv(buf, ARG_INT_FLAGS, flags);
}

static inline void add_str(struct buffer *buf, qstr_t *str)
{
	add_tlv(buf, ARG_STR, str->len, str->name);
}

void stream_llseek_pre(struct assembly_driver *this, struct trace_destination *dest, struct file *file, loff_t offset, int origin, void **context)
{
    PRE_HEADER(OP_LLSEEK);
    add_file(buf, file);
    add_dentry(buf,file->f_dentry);
    add_inode(buf, file->f_dentry->d_inode);
    add_tlv(buf,ARG_LOFFSET,sizeof(loff_t),&offset);
    add_tv(buf,ARG_INT_FLAGS,origin);
    PRE_FOOTER;
}

void stream_llseek_post(struct assembly_driver *this, struct trace_destination *dest, struct file *file, loff_t offset, int origin, loff_t ret, void **context)
{
    int iret;
    POST_HEADER(OP_LLSEEK);
    add_file(buf, file);
    add_dentry(buf,file->f_dentry);
    add_inode(buf, file->f_dentry->d_inode);
    add_tlv(buf,ARG_LOFFSET,sizeof(loff_t),&offset);
    add_tv(buf,ARG_INT_FLAGS,origin);
    /*
     * XXX: this is not correct, but it is necessary to maintain a
     * consistent return value size for the tracefs compiler, it's a temp. fix
     */
    iret = ret;
    add_ret(buf,iret);
    POST_FOOTER;
}

void stream_open_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode, struct file *file, void **context)
{
    PRE_HEADER(OP_OPEN);
    add_file(buf, file);
    add_dentry(buf, file->f_dentry);
    add_inode(buf, inode);
    PRE_FOOTER;
}

void stream_open_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode, struct file *file, int ret, void **context)
{
    POST_HEADER(OP_OPEN);
    add_dentry(buf, file->f_dentry);
    add_inode(buf, inode);
    add_ret(buf, ret);
    POST_FOOTER;
}

static int stream_init(struct assembly_driver *this, void *options)
{
    int err = -ENOMEM;
    ASSERT(this);
    ASSERT(this->ad_type == AD_STREAM);

    this->ad_private_data = KMALLOC(sizeof(struct stream_pd), GFP_KERNEL);
    if (!this->ad_private_data)
	goto out;

    pd(this)->verbosity = (unsigned int)options;
    if (VERBOSITY(this, STR_POST_OP))
	    printk("verbosity POST_OP");

    pd(this)->process_bmap = (char *)get_zeroed_page(GFP_KERNEL);
    if (!pd(this)->process_bmap) {
	KFREE(this->ad_private_data);
	this->ad_private_data = NULL;
	goto out;
    }
    pd(this)->max_process_seen = 0;
    sema_init(&this->ad_odriver_sem, 1);
    err = 0;

  out:
    return err;

}

static void stream_release(struct assembly_driver *this)
{
    ASSERT(this);
    ASSERT(this->ad_type == AD_STREAM);
    if (this->ad_private_data) {
	ASSERT(pd(this)->process_bmap);
	free_page((unsigned long)pd(this)->process_bmap);
	KFREE(this->ad_private_data);
	this->ad_private_data = NULL;
    }
}

static inline void stream_add_header(struct assembly_driver *this, struct trace_destination *dest)
{
    struct stream_header header;
    header.magic = htonl(TRACEFS_STREAM_MAGIC);
    header.version = htonl(VERSION_CODE);
    down(&this->ad_odriver_sem);
    if (dest->td_state != TD_ERROR) {
	if (dest->td_od->od_ops->write(dest->td_od, &header, sizeof(header),OO_KERNEL_BUFFER)
			!= sizeof(header))
		dest->td_state = TD_ERROR;
    }
    up(&this->ad_odriver_sem);
}

void stream_start_trace(struct assembly_driver *this, struct trace_destination *dest)
{
    struct buffer *buf = get_buffer();
    int i;
    op_t tmp = MSG_START;
    stream_add_header(this, dest);
    if (!buf)
	goto out;
    /*
     * We are not using the standard start_pre_trace_message() because
     * we don't want to have the MSG_PNAME message being created for this
     * trace message.
     */

    memcpy(buf->data, &tmp, sizeof(tmp));

    //Set Sequence number for start message to 0
    memset(buf->data+SEQNUM_LOC,0,sizeof(msg_seqnum_t));
    //Initialize sequence_number = 1
    atomic_set(&sequence_number,1);
    //Init the reserve semaphores
    sema_init(&reserved_context_sem,1);
    sema_init(&reserved_buffer_sem,1);

    add_tv(buf, ARG_PID, current->pid);
    add_timestamp(this, buf);
    add_tv(buf, ARG_UID, current->euid);
    add_tv(buf, ARG_GID, current->egid);
    /* uname stuff */
    add_tlv(buf, ARG_SYSNAME, sizeof(system_utsname.sysname), system_utsname.sysname);
    add_tlv(buf, ARG_RELEASE, sizeof(system_utsname.release), system_utsname.release);
    add_tlv(buf, ARG_VERSION, sizeof(system_utsname.version), system_utsname.version);
    add_tlv(buf, ARG_MACHINE, sizeof(system_utsname.machine), system_utsname.machine);

    for (i = 0; msg_desc[i].str_type != NULL; i++) {
	add_tv(buf, ARG_MSG_ID, msg_desc[i].msg_type);
	add_tlv(buf, ARG_MSG_DESC, strlen(msg_desc[i].str_type), msg_desc[i].str_type);
    }
    for (i = 0; arg_desc[i].str_type != NULL; i++) {
	add_tv(buf, ARG_FIELD_ID, arg_desc[i].arg_type);
	add_tv(buf, ARG_FIELD_LEN, arg_desc[i].size);
	add_tlv(buf, ARG_FIELD_DESC, strlen(arg_desc[i].str_type), arg_desc[i].str_type);
    }
    end_trace_message(buf);
    down(&this->ad_odriver_sem);
    if (dest->td_state != TD_ERROR) {
	if (dest->td_od->od_ops->write(dest->td_od, buf->data, buf->ptr - buf->data,OO_KERNEL_BUFFER) != buf->ptr - buf->data)
		dest->td_state = TD_ERROR;
    }
    up(&this->ad_odriver_sem);
    release_buffer(buf);
    if (VERBOSITY(this, STR_SYSCALL_ID)) add_syscall_messages(this, dest);
  out:
    return;
}

void stream_stop_trace(struct assembly_driver *this, struct trace_destination *dest)
{
    struct buffer *buf = get_buffer();
    msg_seqnum_t seqnum = atomic_read(&sequence_number);
    atomic_inc(&sequence_number);
    if (!buf)
		goto out;
    start_pre_trace_message(this, dest, buf, (MSG_STOP));
    add_timestamp(this, buf);
    memcpy(buf->data+SEQNUM_LOC,&seqnum,sizeof(seqnum));
    if (VERBOSITY(this, STR_ARGS)) {
    PRE_FOOTER;
}

static void stream_create_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, int mode, void **context)
{
    PRE_HEADER(OP_CREATE);
    add_dentry(buf, dentry);
    add_flags(buf, mode);
    add_inode(buf, dir);
    PRE_FOOTER;
}
static void stream_create_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, int mode, int ret, void **context)
{
    POST_HEADER(OP_CREATE);
	//During creating a regualar file, add the inode of the new file to add to inode-dentry hash
	//this will be later used by syncpage/ writepage operations to back reference the dentry from
	//the inode
	if(dentry->d_inode)
        add_tv(buf, ARG_CHILD_INUM, dentry->d_inode->i_ino);
    add_dentry(buf, dentry);
    add_flags(buf, mode);
    add_inode(buf, dir);
    add_ret(buf, ret);
    POST_FOOTER;
}

static void stream_lookup_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, void **context)
{
    PRE_HEADER(OP_LOOKUP);
    //Dentry DENTRY_PTR of the child to lookup and the inode ARG_INUM if exists added
    add_dentry(buf, dentry);
    if (dentry->d_inode)
	    add_inode2(buf, dentry->d_inode);
    add_dentry2(buf, dentry->d_parent);
    PRE_FOOTER;
}
static void stream_lookup_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, dentry_t *ret_dentry, void **context)
{
	int ret;
	POST_HEADER(OP_LOOKUP);

	//During lookup, add the inode of the new file to add to inode-dentry hash
    //this will be later used by syncpage/ writepage operations to back reference the dentry from
    //the inode. It is anyway added in add_dentry but for sake of convention add it as CHILD_INUM too.

	if (dentry->d_inode)
        add_tv(buf, ARG_CHILD_INUM, dentry->d_inode->i_ino);

	//Dentry DENTRY_PTR of the child to lookup and the inode ARG_INUM if exists added
	add_dentry(buf, dentry);
   	add_dentry2(buf, dentry->d_parent);
	add_inode2(buf, dir);
	ret = PTR_ERR(ret_dentry);
	add_ret(buf, ret);
	POST_FOOTER;
}

static void stream_unlink_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, void **context)
{
    PRE_HEADER(OP_UNLINK);
    add_dentry(buf, dentry);
    add_inode(buf, dir);
    PRE_FOOTER;
}

static void stream_unlink_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, int ret, void **context)
{
    POST_HEADER(OP_UNLINK);
    add_dentry(buf, dentry);
    add_inode(buf, dir);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_symlink_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, const char *symname, void **context)
{
    PRE_HEADER(OP_SYMLINK);
    add_dentry(buf, dentry);
    add_inode(buf, dir);
    add_tlv(buf, ARG_PATHNAME, strlen(symname), symname);
    PRE_FOOTER;
}

static void stream_symlink_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, const char *symname, int ret, void **context)
{
    POST_HEADER(OP_SYMLINK);
    add_dentry(buf, dentry);
    add_inode(buf, dir);
    add_tlv(buf, ARG_PATHNAME, strlen(symname), symname);
    add_ret(buf, ret);
    POST_FOOTER;
}

static void stream_link_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *old_dentry, struct inode *dir, struct dentry *new_dentry, void **context)
{
    PRE_HEADER(OP_LINK);
    add_dentry(buf, old_dentry);
    add_inode(buf, old_dentry->d_inode);
    PRE_FOOTER;
}
static void stream_link_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *old_dentry, struct inode *dir, struct dentry *new_dentry, int ret, void **context)
{
    POST_HEADER(OP_LINK);
    add_dentry(buf, old_dentry);
    add_inode(buf, old_dentry->d_inode);
    add_tv(buf, ARG_DIR_INUM, dir->i_ino);
    add_dentry2(buf,new_dentry);
    add_ret(buf, ret);
    POST_FOOTER;
}

static void stream_mkdir_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode* dir, struct dentry *dentry, int mode, void **context)
{
    PRE_HEADER(OP_MKDIR);
    add_inode(buf, dir);
    add_dentry(buf, dentry);
    add_flags(buf, mode);
    PRE_FOOTER;
}
static void stream_mkdir_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, int ret, void **context)
{
    POST_HEADER(OP_MKDIR);
	//During creating a directory file, add the inode of the new directory to add to inode-dentry hash
    //this will be later used by syncpage/ writepage operations to back reference the dentry from
    //the inode
	if (dentry->d_inode)
	    add_tv(buf, ARG_CHILD_INUM, dentry->d_inode->i_ino);

    add_inode(buf, dir);
    add_dentry(buf, dentry);
    add_flags(buf, mode);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_rmdir_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, void **context)
{
    PRE_HEADER(OP_RMDIR);
    add_inode(buf, dir);
    add_dentry(buf, dentry);
    PRE_FOOTER;
}
static void stream_rmdir_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int ret, void **context)
{
    POST_HEADER(OP_RMDIR);
    add_inode(buf, dir);
    add_dentry(buf, dentry);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_mknod_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, unsigned long dev, void **context)
{
    int idev;
    PRE_HEADER(OP_MKNOD);
    idev = (int)dev;
    add_inode(buf, dir);
    add_dentry(buf, dentry);
    add_flags(buf, mode);
    add_tv(buf,ARG_ULONG,idev);
    PRE_FOOTER;
}
static void stream_mknod_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, unsigned long dev, int ret, void **context)
{
    int idev;
    POST_HEADER(OP_MKNOD);
    idev = (int)dev;
	//During creating a device, add the inode of the new device to add to inode-dentry hash
    //this will be later used by syncpage/ writepage operations to back reference the dentry from
    //the inode
    if (dentry->d_inode)
        add_tv(buf, ARG_CHILD_INUM, dentry->d_inode->i_ino);

    add_inode(buf, dir);
    add_dentry(buf, dentry);
    add_flags(buf, mode);
    add_tv(buf,ARG_ULONG,idev);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_rename_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry, void **context)
{
    PRE_HEADER(OP_RENAME);
    add_inode(buf, old_dir);
    add_dentry(buf, old_dentry);
    add_inode2(buf, new_dir);
    add_dentry2(buf, new_dentry);
    PRE_FOOTER;
}
static void stream_rename_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry, int ret, void **context)
{
    POST_HEADER(OP_RENAME);
    add_inode(buf, old_dir);
    add_dentry(buf, old_dentry);
    add_inode2(buf, new_dir);
    add_dentry2(buf, new_dentry);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_readlink_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, char *buffer, int bufsiz, void **context)
{
    PRE_HEADER(OP_READLINK);
    add_dentry(buf, dentry);
    add_flags(buf, bufsiz);
    PRE_FOOTER;
}
static void stream_readlink_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, char *buffer, int bufsiz, int ret, void **context)
{
    POST_HEADER(OP_READLINK);
    add_dentry(buf, dentry);
    if (ret == 0) {
	    add_tlv(buf, ARG_STR, strlen(buffer), buffer);
    }
    add_flags(buf, bufsiz);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_follow_link_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct nameidata *nd, void **context)
{
    PRE_HEADER(OP_FOLLOW_LINK);
    /* XXX nameidata */
    add_dentry(buf, dentry);
    PRE_FOOTER;
}
static void stream_follow_link_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct nameidata *nd, int ret, void **context)
{
    POST_HEADER(OP_FOLLOW_LINK);
    add_dentry(buf, dentry);
    /* XXX nameidata */
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_permission_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode,int mask, void **context)
{
    PRE_HEADER(OP_PERMISSION);
    add_inode(buf, inode);
    add_flags(buf, mask);
    PRE_FOOTER;
}
static void stream_permission_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode,int mask, int ret, void **context)
{
    POST_HEADER(OP_PERMISSION);
    add_inode(buf, inode);
    add_flags(buf, mask);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_inode_revalidate_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, void **context)
{
    PRE_HEADER(OP_INODE_REVALIDATE);
    add_dentry(buf, dentry);
    PRE_FOOTER;
}
static void stream_inode_revalidate_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, int ret, void **context)
{
    POST_HEADER(OP_INODE_REVALIDATE);
    add_dentry(buf, dentry);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_setattr_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct iattr *ia, void **context)
{
    PRE_HEADER(OP_SETATTR);
    add_dentry(buf, dentry);
    add_tlv(buf,ARG_IATTR,sizeof(struct iattr),ia);
    PRE_FOOTER;
}
static void stream_setattr_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct iattr *ia, int ret, void **context)
{
    POST_HEADER(OP_SETATTR);
    add_dentry(buf, dentry);
    add_tlv(buf,ARG_IATTR,sizeof(struct iattr),ia);
    add_ret(buf, ret);
    POST_FOOTER;
}
/* file ops */
static void stream_read_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, char *buffer, size_t count, loff_t *ppos, void **context)
{
    PRE_HEADER(OP_READ);
    add_file(buf, file);
    add_tv(buf, ARG_ULONG, count);
    add_tv(buf, ARG_LOFFSET, *ppos);
    PRE_FOOTER;
}

/*
 * This is second stage for read/write calls that may trace data.
 * This function exists to save code duplication
 */


static inline void add_message(struct assembly_driver *this, struct trace_destination *dest, int ret, struct buffer *buf, const char *buffer, size_t count,op_t op,int is_user_buffer)
{
	char log_data = 0; /* ugly state variable to save some computation */
	size_t extra = 0;

	/* The following code is copied from POST_FOOTER and modified to support data tracing */
	if (ret > 0 && VERBOSITY(this, STR_ARGS) &&
    VERBOSITY(this, STR_DATA))
  {
		if (VERBOSITY(this, STR_DATA_CKSUM))
    {
		  md5_state_t md5_state;

		  add_tl(buf, ARG_CKSUM, MD5_CKSUM_LEN);
		  md5_init(&md5_state);
		  md5_append(&md5_state, buffer, count);
		  md5_finish(&md5_state, buf->ptr);
		  buf->ptr += MD5_CKSUM_LEN;
		}
    else
    {
			log_data = 1;
			add_tl(buf, ARG_DATA, count);
			/* "fake" the buf->data so that end_trace_message puts the right value
			 * in the length field
			 */
			buf->ptr += count;
			extra = count;
		}
	}


	end_trace_message(buf);
	down(&this->ad_odriver_sem);
	if (dest->td_state != TD_ERROR)
  {
		if (dest->td_od->od_ops->write(dest->td_od,
                  buf->data, buf->ptr - buf->data - extra,
                  is_user_buffer) != buf->ptr - buf->data - extra)
    {
			  dest->td_state = TD_ERROR;

    }

		if (log_data && dest->td_state != TD_ERROR)
    {
			if (dest->td_od->od_ops->write(dest->td_od,
        (char *)buffer, count,is_user_buffer) != count)
      {
				dest->td_state = TD_ERROR;

      }

		}
	}
	up(&this->ad_odriver_sem);
}

static void stream_read_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, char *buffer, size_t count, loff_t *ppos, ssize_t ret, void **context)
{
    POST_HEADER(OP_READ);
		add_file(buf, file);
		add_tv(buf, ARG_ULONG, count);
		add_tv(buf, ARG_LOFFSET, *ppos);
		add_ret(buf, ret);
	}
        if(!VERBOSITY(this,STR_DATA_WITHOUT_READ))
                add_message(this, dest, ret, buf, buffer, count,OP_READ,OO_KERNEL_BUFFER);
	release_buffer(buf);
    }
out:
	return;

}
static void stream_write_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, const char *buffer, size_t count, loff_t *ppos, void **context)
{
    PRE_HEADER(OP_WRITE);
    add_file(buf, file);
    add_tv(buf, ARG_ULONG, count);
    add_tv(buf, ARG_LOFFSET, *ppos);
    PRE_FOOTER;
}

static void stream_write_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, const char *buffer, size_t count, loff_t *ppos, ssize_t ret, void **context)
{
	POST_HEADER(OP_WRITE);
			add_file(buf, file);
			add_tv(buf, ARG_ULONG, count);
			add_tv(buf, ARG_LOFFSET, *ppos);
			add_ret(buf, ret);
		}
	  add_message(this, dest, ret, buf, buffer, count,OP_WRITE,OO_USER_BUFFER);
		release_buffer(buf);
	}
out:
	return;
}

static void stream_readdir_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, void *dirent, filldir_t filldir, int count, void **context)
{
    PRE_HEADER(OP_READDIR);
    add_file(buf, file);
    add_flags(buf, count);
    PRE_FOOTER;
}
static void stream_readdir_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, void *firent, filldir_t filldir, int count, int ret, void **context)
{
    POST_HEADER(OP_READDIR);
    add_file(buf, file);
    add_flags(buf, count);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_poll_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, poll_table *wait, void **context)
{
    PRE_HEADER(OP_POLL);
    add_file(buf, file);
    /* XXX poll_table */
    PRE_FOOTER;
}
static void stream_poll_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, poll_table *wait, int ret, void **context)
{
    POST_HEADER(OP_POLL);
    add_file(buf, file);
    /* XXX poll_table */
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_ioctl_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg, void **context)
{
    PRE_HEADER(OP_IOCTL);
    add_inode(buf, inode);
    add_file(buf, file);
    add_flags(buf, cmd);
    add_tv(buf, ARG_ULONG, arg);
    PRE_FOOTER;
}
static void stream_ioctl_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg, int ret, void **context)
{
    POST_HEADER(OP_IOCTL);
    add_inode(buf, inode);
    add_file(buf, file);
    add_flags(buf, cmd);
    add_tv(buf, ARG_ULONG, arg);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_mmap_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, vm_area_t *vma, void **context)
{
    PRE_HEADER(OP_MMAP);
    add_file(buf, file);
    /* XXX vm_area_t */
    PRE_FOOTER;
}
static void stream_mmap_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, vm_area_t *vma, int ret, void **context)
{
    POST_HEADER(OP_MMAP);
    add_file(buf, file);
    /* XXX vm_area_t */
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_flush_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, void **context)
{
    PRE_HEADER(OP_FLUSH);
    add_file(buf, file);
    PRE_FOOTER;
}
static void stream_flush_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, int ret, void **context)
{
    POST_HEADER(OP_FLUSH);
    add_file(buf, file);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_release_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, void **context)
{
    PRE_HEADER(OP_RELEASE);
    add_inode(buf, inode);
    add_file(buf, file);
    PRE_FOOTER;
}
static void stream_release_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, int ret, void **context)
{
    POST_HEADER(OP_RELEASE);
    add_inode(buf, inode);
    add_file(buf, file);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_fsync_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, dentry_t *dentry, int datasync, void **context)
{
    PRE_HEADER(OP_FSYNC);
    add_file(buf, file);
    add_dentry(buf, dentry);
    add_flags(buf, datasync);
    PRE_FOOTER;
}
static void stream_fsync_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, dentry_t *dentry, int datasync, int ret, void **context)
{
    POST_HEADER(OP_FSYNC);
    add_file(buf, file);
    add_dentry(buf, dentry);
    add_flags(buf, datasync);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_fasync_pre(struct assembly_driver *this, struct trace_destination *dest, int fd, file_t *file, int flag, void **context)
{
    PRE_HEADER(OP_FASYNC);
    add_tv(buf, ARG_FD, fd);
    add_file(buf, file);
    add_flags(buf, flag);
    PRE_FOOTER;
}
static void stream_fasync_post(struct assembly_driver *this, struct trace_destination *dest, int fd, file_t *file, int flag, int ret, void **context)
{
    POST_HEADER(OP_FASYNC);
    add_file(buf, file);
    add_flags(buf, flag);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_lock_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, int cmd, struct file_lock *fl, void **context)
{
    PRE_HEADER(OP_LOCK);
    add_file(buf, file);
    add_flags(buf, cmd);
    /* XXX: file_lock */
    PRE_FOOTER;
}
static void stream_lock_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, int cmd, struct file_lock *fl, int ret, void **context)
{
    POST_HEADER(OP_LOCK);
    add_file(buf, file);
    add_flags(buf, cmd);
    /* XXX: file_lock */
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_read_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    PRE_HEADER(OP_READ_INODE);
    add_inode(buf, inode);
    PRE_FOOTER;
}
static void stream_write_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, int sync, void **context)
{
    PRE_HEADER_NO_POST(OP_WRITE_INODE);
    add_inode(buf, inode);
    add_flags(buf, sync);
    PRE_FOOTER;
}
static void stream_put_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    PRE_HEADER_NO_POST(OP_PUT_INODE);
    add_inode(buf, inode);
    PRE_FOOTER;
}
static void stream_delete_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    PRE_HEADER_NO_POST(OP_DELETE_INODE);
    add_inode(buf, inode);
    PRE_FOOTER;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static void stream_statfs_pre(struct assembly_driver *this, struct trace_destination *dest, struct statfs *buffer, void **context)
#else
static void stream_statfs_pre(struct assembly_driver *this, struct trace_destination *dest, struct kstatfs *buffer, void **context)
#endif
{
    PRE_HEADER(OP_STATFS);
    /* XXX statfs struct */
    PRE_FOOTER;
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static void stream_statfs_post(struct assembly_driver *this, struct trace_destination *dest, struct statfs *buffer, int ret, void **context)
#else
static void stream_statfs_post(struct assembly_driver *this, struct trace_destination *dest, struct kstatfs *buffer, int ret, void **context)
#endif
{
    POST_HEADER(OP_STATFS);
    /* XXX statfs struct */
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_clear_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    PRE_HEADER_NO_POST(OP_CLEAR_INODE);
    add_inode(buf, inode);
    PRE_FOOTER;
}
static void stream_d_revalidate_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, int flags, void **context)
{
    PRE_HEADER(OP_D_REVALIDATE);
    add_dentry(buf, dentry);
    add_flags(buf, flags);
    PRE_FOOTER;
}
static void stream_d_revalidate_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, int flags, int ret, void **context)
{
    POST_HEADER(OP_D_REVALIDATE);
    add_dentry(buf, dentry);
    add_flags(buf, flags);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_d_hash_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *name, void **context)
{
    PRE_HEADER(OP_D_HASH);
    add_dentry(buf, dentry);
    add_str(buf, name);
    PRE_FOOTER;
}
static void stream_d_hash_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *name, int ret, void **context)
{
    POST_HEADER(OP_D_HASH);
    add_dentry(buf, dentry);
    add_str(buf, name);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_d_compare_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *a, qstr_t *b, void **context)
{
    PRE_HEADER(OP_D_COMPARE);
    add_dentry(buf, dentry);
    add_str(buf, a);
    add_str(buf, b);
    PRE_FOOTER;
}
static void stream_d_compare_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *a, qstr_t *b, int ret, void **context)
{
    POST_HEADER(OP_D_COMPARE);
    add_dentry(buf, dentry);
    add_str(buf, a);
    add_str(buf, b);
    add_ret(buf, ret);
    POST_FOOTER;
}
static void stream_d_delete_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, void **context)
{
    PRE_HEADER(OP_D_DELETE);
    add_dentry(buf, dentry);
    PRE_FOOTER;
}
static void stream_d_delete_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, int ret, void **context)
{
    POST_HEADER(OP_D_DELETE);
    add_dentry(buf, dentry);
    add_ret(buf, ret);
    POST_FOOTER;
}

static void stream_d_release_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, void **context)
{
    PRE_HEADER(OP_D_RELEASE);
    add_dentry(buf, dentry);
    PRE_FOOTER;
}

static void stream_d_release_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, void **context)
{
    POST_HEADER(OP_D_RELEASE);
    add_dentry(buf, dentry);
    POST_FOOTER;
}

static void stream_d_iput_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, inode_t *inode, void **context)
{
    PRE_HEADER_NO_POST(OP_D_IPUT);
    add_dentry(buf, dentry);
    add_inode(buf, inode);
    PRE_FOOTER;
}

static void stream_readpage_pre(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, void **context)
{
	PRE_HEADER(OP_READPAGE);
	add_file(buf, file);
	add_page(buf, page);
	PRE_FOOTER;
}

static void stream_readpage_post(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, int ret, void **context)
{
	char* buffer;

	POST_HEADER(OP_READPAGE);
			add_file(buf, file);
			add_page_with_data(buf, page);
			add_ret(buf, ret);
		}
		buffer = kmap(page);
                if(!VERBOSITY(this,STR_DATA_WITHOUT_READ))
            		add_message(this, dest, 1, buf, buffer, PAGE_SIZE,OP_READPAGE,OO_KERNEL_BUFFER);
		kunmap(page);
		release_buffer(buf);
	}
out:
	return;
}

static void stream_writepage_pre(struct assembly_driver *this, struct trace_destination *dest, struct page *page, void **context)
{
	PRE_HEADER_NOFS(OP_WRITEPAGE);
	add_page_with_data(buf, page);
	add_inode(buf, page->mapping->host);
	PRE_FOOTER;
}

static void stream_writepage_post(struct assembly_driver *this, struct trace_destination *dest, struct page *page, int ret, void **context)
{
	POST_HEADER_NOFS(OP_WRITEPAGE);
	add_page_with_data(buf, page);
			add_inode(buf, page->mapping->host);
			add_ret(buf, ret);
		}
	  add_message(this, dest, 1, buf, kmap(page), PAGE_SIZE,OP_WRITEPAGE,OO_KERNEL_BUFFER);
		kunmap(page);
                if(buf != &reserved_buffer)
                        release_buffer(buf);
                else
		        up(&reserved_buffer_sem);
		}
out:
	return;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static void stream_readpages_pre(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct address_space* mapping,struct list_head *pages, void **context)
{
	PRE_HEADER(OP_READPAGES);
	add_file(buf, file);
        add_pages(buf, pages);
	PRE_FOOTER;
}

static void stream_readpages_post(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct address_space* mapping,struct list_head *pages, int ret, void **context)
{
	char* buffer;
        page_t* page;
	page_t* first_page;
        struct list_head* list;

	POST_HEADER(OP_READPAGES);
			add_file(buf, file);
			add_pages_with_data(buf, pages);
			add_ret(buf, ret);
		}
                first_page = NULL;
                //Assume circular list
                list_for_each(list,pages) {
                        page = list_entry(pages,page_t,lru);
                        if(first_page == page)
			  break;
			else if(first_page == NULL)
			  first_page = page;
         		buffer = kmap(page);
			add_message(this, dest, 1, buf, buffer, PAGE_SIZE,OP_READPAGES,OO_KERNEL_BUFFER);
	        	kunmap(page);
                }
                release_buffer(buf);
	}
out:
	return;
}
#endif //#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

static void stream_prepare_write_pre(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, void **context)
{
	PRE_HEADER(OP_PREPARE_WRITE);
	add_file(buf, file);
	add_page(buf, page);
	add_tv(buf, ARG_WRITE_FROM, from);
	add_tv(buf, ARG_WRITE_TO, to);
	PRE_FOOTER;
}
void stream_prepare_write_post(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, int ret, void **context)
{
	POST_HEADER(OP_PREPARE_WRITE);
	add_file(buf, file);
	add_page(buf, page);
	add_tv(buf, ARG_WRITE_FROM, from);
	add_tv(buf, ARG_WRITE_TO, to);
	add_ret(buf, ret);
	POST_FOOTER;
}
void stream_commit_write_pre(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, void **context)
{
	PRE_HEADER(OP_COMMIT_WRITE);
	add_file(buf, file);
	add_page_with_data(buf, page);
	add_tv(buf, ARG_WRITE_FROM, from);
	add_tv(buf, ARG_WRITE_TO, to);
	PRE_FOOTER;
}
void stream_commit_write_post(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, int ret, void **context)
{
	/*POST_HEADER(OP_COMMIT_WRITE);
	add_file(buf, file);
	add_page(buf, page);
	add_tv(buf, ARG_WRITE_FROM, from);
	add_tv(buf, ARG_WRITE_TO, to);
	add_ret(buf, ret);
	POST_FOOTER;*/
	POST_HEADER(OP_COMMIT_WRITE);
			add_file(buf, file);
			add_page_with_data(buf, page);
			add_tv(buf, ARG_WRITE_FROM, from);
			add_tv(buf, ARG_WRITE_TO, to);
			add_inode(buf, page->mapping->host);
			add_ret(buf, ret);
		}
        	add_message(this, dest, 1, buf, kmap(page), PAGE_SIZE,OP_COMMIT_WRITE,OO_KERNEL_BUFFER);
		kunmap(page);
		release_buffer(buf);
	}
out:
	return;
}

void stream_sync_page_pre(struct assembly_driver *this,struct trace_destination *dest,page_t* page,void **context)
{
  PRE_HEADER(OP_SYNC_PAGE);
  add_page(buf,page);
  add_inode(buf,page->mapping->host);

  PRE_FOOTER;
}

void stream_sync_page_post(struct assembly_driver *this,struct trace_destination *dest,page_t* page,int ret,void **context)
{
  POST_HEADER(OP_SYNC_PAGE);
  add_page(buf,page);
  add_inode(buf,page->mapping->host);
  add_ret(buf,ret);
  POST_FOOTER;
}

static struct assembly_ops stream_ad_ops = {
init: stream_init,
release: stream_release,
ao_llseek_pre: stream_llseek_pre,
ao_llseek_post: stream_llseek_post,
ao_create_pre: stream_create_pre,
ao_create_post: stream_create_post,
ao_link_pre: stream_link_pre,
ao_link_post: stream_link_post,
ao_lookup_pre: stream_lookup_pre,
ao_lookup_post: stream_lookup_post,
ao_unlink_pre: stream_unlink_pre,
ao_unlink_post: stream_unlink_post,
ao_symlink_pre: stream_symlink_pre,
ao_symlink_post: stream_symlink_post,
ao_mkdir_pre: stream_mkdir_pre,
ao_mkdir_post: stream_mkdir_post,
ao_inode_revalidate_pre: stream_inode_revalidate_pre,
ao_inode_revalidate_post: stream_inode_revalidate_post,
ao_permission_pre: stream_permission_pre,
ao_permission_post: stream_permission_post,
ao_follow_link_pre: stream_follow_link_pre,
ao_follow_link_post: stream_follow_link_post,
ao_readlink_pre: stream_readlink_pre,
ao_readlink_post: stream_readlink_post,
ao_rename_pre: stream_rename_pre,
ao_rename_post: stream_rename_post,
ao_mknod_pre: stream_mknod_pre,
ao_mknod_post: stream_mknod_post,
ao_rmdir_pre: stream_rmdir_pre,
ao_rmdir_post: stream_rmdir_post,
ao_setattr_pre: stream_setattr_pre,
ao_setattr_post: stream_setattr_post,
ao_read_pre : stream_read_pre,
ao_read_post : stream_read_post,
ao_write_pre : stream_write_pre,
ao_write_post : stream_write_post,
ao_readdir_pre : stream_readdir_pre,
ao_readdir_post : stream_readdir_post,
ao_poll_pre : stream_poll_pre,
ao_poll_post : stream_poll_post,
ao_ioctl_pre : stream_ioctl_pre,
ao_ioctl_post : stream_ioctl_post,
ao_mmap_pre : stream_mmap_pre,
ao_mmap_post : stream_mmap_post,
ao_open_pre: stream_open_pre,
ao_open_post: stream_open_post,
ao_flush_pre : stream_flush_pre,
ao_flush_post : stream_flush_post,
ao_release_pre : stream_release_pre,
ao_release_post : stream_release_post,
ao_fsync_pre : stream_fsync_pre,
ao_fsync_post : stream_fsync_post,
ao_fasync_pre : stream_fasync_pre,
ao_fasync_post : stream_fasync_post,
ao_lock_pre : stream_lock_pre,
ao_lock_post : stream_lock_post,
/* super ops */
ao_read_inode_pre : stream_read_inode_pre,
ao_write_inode_pre : stream_write_inode_pre,
ao_put_inode_pre : stream_put_inode_pre,
ao_delete_inode_pre : stream_delete_inode_pre,
ao_statfs_pre : stream_statfs_pre,
ao_statfs_post : stream_statfs_post,
ao_clear_inode_pre : stream_clear_inode_pre,
/* dentry ops */
ao_d_revalidate_pre : stream_d_revalidate_pre,
ao_d_revalidate_post : stream_d_revalidate_post,
ao_d_hash_pre : stream_d_hash_pre,
ao_d_hash_post : stream_d_hash_post,
ao_d_compare_pre : stream_d_compare_pre,
ao_d_compare_post : stream_d_compare_post,
ao_d_delete_pre : stream_d_delete_pre,
ao_d_delete_post : stream_d_delete_post,
ao_d_release_pre : stream_d_release_pre,
ao_d_release_post : stream_d_release_post,
ao_d_iput_pre : stream_d_iput_pre,
ao_readpage_pre: stream_readpage_pre,
ao_readpage_post: stream_readpage_post,
ao_writepage_pre: stream_writepage_pre,
ao_writepage_post: stream_writepage_post,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
ao_readpages_pre: stream_readpages_pre,
ao_readpages_post: stream_readpages_post,
#endif //#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
ao_prepare_write_pre: stream_prepare_write_pre,
ao_prepare_write_post: stream_prepare_write_post,
ao_commit_write_pre: stream_commit_write_pre,
ao_commit_write_post: stream_commit_write_post,
ao_sync_page_pre: stream_sync_page_pre,
ao_sync_page_post: stream_sync_page_post,
ao_start_trace: stream_start_trace,
ao_stop_trace: stream_stop_trace
};

struct assembly_driver *create_stream_driver()
{
    struct assembly_driver *tmp = (struct assembly_driver *)KMALLOC(sizeof(struct assembly_driver), GFP_KERNEL);

    if (tmp) {
	tmp->ad_type = AD_STREAM;
	tmp->ad_ops = &stream_ad_ops;
	tmp->ad_next = NULL;
	tmp->ad_private_data = NULL;
    }
    return tmp;


}


void destroy_stream_driver(struct assembly_driver *stream)
{
    ASSERT(stream);
    ASSERT(stream->ad_type == AD_STREAM);
    ASSERT(stream->ad_private_data == NULL);

    /* only free as of now */
    KFREE(stream);
}
