/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * od_buffer.c
 * output driver for file based output
 */

#include "tracer.h"
#include "od_buffer.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
/*
 * internal structure for this driver
 */

struct od_buffer_data {
	char *buf;     /* buffer */
	int remaining; /* remaining bytes in buffer */
	int buf_size;
};

/**
 * Destroys the file driver object.
 * Make sure that release function has already been called
 * @param output driver to be destroyed
 */

void destroy_buffer_driver(struct output_driver *od)
{
	ASSERT(od);
	ASSERT(od->od_private_data == NULL);
	/* nothing to do, as of now */
	KFREE(od);
}

/**
 * Initializes the file driver object.
 * @param File driver object to initialize
 * @param <code>struct od_buffer_params</code> type initialization parameters
 * @see struct od_buffer_params
 */

static int bdriver_init(struct output_driver *this, void *params, struct output_driver *next, int level)
{
	int ret = -ENOMEM;

	struct od_buffer_data *fdata;
	struct od_buffer_params *fp = (struct od_buffer_params *)params;
	ASSERT(this->od_private_data == NULL);
	if (!next || !params) {
		ret = -EINVAL;
		goto out;
	}

	print_entry_location();

	fdata = this->od_private_data = KMALLOC(sizeof(struct od_buffer_data), GFP_KERNEL);
	if (!fdata)
		goto out;
	memset(this->od_private_data, 0, sizeof(struct od_buffer_data));
	fdata->buf_size = fp->buf_size;
	if (fdata->buf_size == 0)
		fdata->buf_size = next->od_ops->getblocksize(next);
	/* allocate the buffer */
	fdata->remaining = fdata->buf_size;
	printk("asking for %d\n", fdata->buf_size);
	fdata->buf = (char *)vmalloc(fdata->buf_size);
	if (!fdata->buf)
		goto out_free;
	/* all good! */
	this->od_next = next;
	ret = 0;
	goto out;

out_free:
	KFREE(fdata);
	this->od_private_data = NULL;

out:
	print_exit_status(ret);
	return ret;
}

/**
 * Releases the private data.
 * Flush should be called before this. Also causes the chained drivers
 * to be released.
 * @param driver to release
 */
void buffer_release(struct output_driver *this)
{
	struct od_buffer_data *fdata;
	ASSERT(this);
	ASSERT(this->od_type == OD_BUFFER);
	fdata = this->od_private_data;
	/* free the buffer, if any */
	vfree(fdata->buf);
	KFREE(fdata);
	this->od_private_data = NULL;
}

/**
 * flushes the interbal buffer, if any.
 * @param object to operate on
 * @return Number of bytes written.
 */
int buffer_flush(struct output_driver *this)
{
	struct od_buffer_data *fdata;
	int ret = 0;

	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_BUFFER);
	fdata = this->od_private_data;

	if (fdata->remaining < fdata->buf_size) {
	    if (this->od_next)
		ret = this->od_next->od_ops->write(this->od_next, fdata->buf, fdata->buf_size - fdata->remaining,OO_KERNEL_BUFFER);
	    fdata->remaining = fdata->buf_size;
	}
	if (ret >= 0) {
		ret = this->od_next->od_ops->flush(this->od_next);
	}
	print_exit_status(ret);
	return ret;
}

/**
 * Writes the data to output.
 * If internal buffering is enabled, data is written to buffer
 * till the buffer runs out.
 * @param object to operate on
 * @param buffer to write out
 * @param number of bytes to write
 * @param whether the buffer is a user-level buffer (OO_USER_BUFFER)
 * @returns number of bytes written
 */

int buffer_write(struct output_driver *this, void *buf, int len,int is_user_buffer)
{
	int ret = 0;
	int total;
	int junk;
	struct od_buffer_data *fdata;
	char* tmp;
	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_BUFFER);
	fdata = this->od_private_data;

	tmp = (char *)buf;
	total = len;
	while(fdata->remaining < len) {
		len -= fdata->remaining;
		if(is_user_buffer == OO_KERNEL_BUFFER)
		  memcpy(fdata->buf + fdata->buf_size - fdata->remaining, tmp, fdata->remaining);
		else
		  junk = copy_from_user(fdata->buf + fdata->buf_size - fdata->remaining, tmp, fdata->remaining);
		tmp += fdata->remaining;
		ret = this->od_next->od_ops->write(this->od_next, fdata->buf, fdata->buf_size,is_user_buffer);
		if (ret < 0)
			goto out;
		fdata->remaining = fdata->buf_size;

	}
	if (len > 0) {
         	if(is_user_buffer == OO_KERNEL_BUFFER)
		  memcpy(fdata->buf + fdata->buf_size - fdata->remaining, tmp, len);
		else
		  junk = copy_from_user(fdata->buf + fdata->buf_size - fdata->remaining, tmp, len);
		fdata->remaining -= len;
	}
	ret = total;

out:
	print_exit_status(ret);
	return ret;
}

/**
 * Returns the preferred block size
 * @param object to operate on
 * @returns block size
 */
int buffer_getblocksize(struct output_driver *this)
{
    struct od_buffer_data *fdata;
    ASSERT(this);
    ASSERT(this->od_type == OD_BUFFER);
    fdata = this->od_private_data;
    return fdata->buf_size;
}

static struct output_ops buffer_od_ops = {
    write: buffer_write,
    flush: buffer_flush,
    getblocksize: buffer_getblocksize,
    init: bdriver_init,
    release: buffer_release
};

/**
 * Creates a file driver object.
 * @return A file output driver object with function pointers initialised
 */

struct output_driver* create_buffer_driver()
{
    struct output_driver *driver = (struct output_driver *)KMALLOC(sizeof(struct output_driver), GFP_KERNEL);

    if (driver) {
	driver->od_ops = &buffer_od_ops;
	driver->od_type = OD_BUFFER;
	driver->od_private_data = NULL;
	driver->od_next = NULL;
    }
    return driver;
}


